from enum import Enum
from typing import Iterable, Union, List
import re
from datetime import timedelta

# use lxml if available (should be present as a dependency of the compal library)
from prometheus_client.metrics_core import (
    InfoMetricFamily,
    CounterMetricFamily,
    GaugeMetricFamily,
    StateSetMetricFamily,
)

try:
    from lxml import etree
except ImportError:
    import elementtree.ElementTree as etree

from prometheus_client import Metric
from compal.functions import GetFunction as GET


class XmlMetricsExtractor:
    def __init__(self, functions: Union[int, List[int]]):
        if type(functions) is list:
            self._functions = functions
        else:
            self._functions = [functions]
        self.parser = etree.XMLParser()

    @property
    def functions(self) -> List[int]:
        """
        Connect Box getter.xml function(s) this metrics extractor is working on
        :return:
        """
        return self._functions

    def extract(self, raw_xmls: List[bytes]) -> Iterable[Metric]:
        """
        Returns metrics given raw XML responses corresponding to the functions returned in the `functions` property.
        :param raw_xmls:
        :return:
        """
        raise NotImplementedError


class DownstreamStatusExtractor(XmlMetricsExtractor):
    def __init__(self):
        super(DownstreamStatusExtractor, self).__init__(
            [GET.DOWNSTREAM_TABLE, GET.SIGNAL_TABLE]
        )

    def extract(self, raw_xmls: List[bytes]) -> Iterable[Metric]:
        assert len(raw_xmls) == 2
        downstream_table_xml = raw_xmls[0]
        signal_table_xml = raw_xmls[1]

        # DOWNSTREAM_TABLE gives us frequency, power level, SNR and RxMER per channel
        root = etree.fromstring(downstream_table_xml, parser=self.parser)
        assert root.tag == "downstream_table"

        CHANNEL_ID = "channel_id"
        ds_frequency = GaugeMetricFamily(
            "connectbox_downstream_frequency",
            "Downstream channel frequency",
            unit="hz",
            labels=[CHANNEL_ID],
        )
        ds_power_level = GaugeMetricFamily(
            "connectbox_downstream_power_level",
            "Downstream channel power level",
            unit="dbmv",
            labels=[CHANNEL_ID],
        )
        ds_snr = GaugeMetricFamily(
            "connectbox_downstream_snr",
            "Downstream channel signal-to-noise ratio (SNR)",
            unit="db",
            labels=[CHANNEL_ID],
        )
        ds_rxmer = GaugeMetricFamily(
            "connectbox_downstream_rxmer",
            "Downstream channel receive modulation error ratio (RxMER)",
            unit="db",
            labels=[CHANNEL_ID],
        )
        for channel in root.findall("downstream"):
            channel_id = channel.find("chid").text
            frequency = int(channel.find("freq").text)
            power_level = int(channel.find("pow").text)
            snr = int(channel.find("snr").text)
            rxmer = float(channel.find("RxMER").text)

            labels = [channel_id.zfill(2)]
            ds_frequency.add_metric(labels, frequency)
            ds_power_level.add_metric(labels, power_level)
            ds_snr.add_metric(labels, snr)
            ds_rxmer.add_metric(labels, rxmer)
        yield from [ds_frequency, ds_power_level, ds_snr, ds_rxmer]

        # SIGNAL_TABLE gives us unerrored, corrected and uncorrectable errors per channel
        root = etree.fromstring(signal_table_xml, parser=self.parser)
        assert root.tag == "signal_table"

        ds_unerrored_codewords = CounterMetricFamily(
            "connectbox_downstream_codewords_unerrored",
            "Unerrored downstream codewords",
            labels=[CHANNEL_ID],
        )
        ds_correctable_codewords = CounterMetricFamily(
            "connectbox_downstream_codewords_corrected",
            "Corrected downstream codewords",
            labels=[CHANNEL_ID],
        )
        ds_uncorrectable_codewords = CounterMetricFamily(
            "connectbox_downstream_codewords_uncorrectable",
            "Uncorrectable downstream codewords",
            labels=[CHANNEL_ID],
        )
        for channel in root.findall("signal"):
            channel_id = channel.find("dsid").text
            unerrored = int(channel.find("unerrored").text)
            correctable = int(channel.find("correctable").text)
            uncorrectable = int(channel.find("uncorrectable").text)

            labels = [channel_id.zfill(2)]
            ds_unerrored_codewords.add_metric(labels, unerrored)
            ds_correctable_codewords.add_metric(labels, correctable)
            ds_uncorrectable_codewords.add_metric(labels, uncorrectable)
        yield from [
            ds_unerrored_codewords,
            ds_correctable_codewords,
            ds_uncorrectable_codewords,
        ]


class UpstreamStatusExtractor(XmlMetricsExtractor):
    def __init__(self):
        super(UpstreamStatusExtractor, self).__init__(GET.UPSTREAM_TABLE)

    def extract(self, raw_xmls: List[bytes]) -> Iterable[Metric]:
        assert len(raw_xmls) == 1
        upstream_table_xml = raw_xmls[0]

        CHANNEL_ID = "channel_id"
        TIMEOUT_TYPE = "timeout_type"

        us_frequency = GaugeMetricFamily(
            "connectbox_upstream_frequency",
            "Upstream channel frequency",
            unit="hz",
            labels=[CHANNEL_ID],
        )
        us_power_level = GaugeMetricFamily(
            "connectbox_upstream_power_level",
            "Upstream channel power level",
            unit="dbmv",
            labels=[CHANNEL_ID],
        )
        us_symbol_rate = GaugeMetricFamily(
            "connectbox_upstream_symbol_rate",
            "Upstream channel symbol rate",
            unit="ksps",
            labels=[CHANNEL_ID],
        )
        us_timeouts = CounterMetricFamily(
            "connectbox_upstream_timeouts",
            "Upstream channel timeout occurrences",
            labels=[CHANNEL_ID, TIMEOUT_TYPE],
        )
        root = etree.fromstring(upstream_table_xml, parser=self.parser)
        assert root.tag == "upstream_table"
        for channel in root.findall("upstream"):
            channel_id = channel.find("usid").text

            frequency = int(channel.find("freq").text)
            power_level = float(channel.find("power").text)
            symbol_rate = float(channel.find("srate").text)
            t1_timeouts = int(channel.find("t1Timeouts").text)
            t2_timeouts = int(channel.find("t2Timeouts").text)
            t3_timeouts = int(channel.find("t3Timeouts").text)
            t4_timeouts = int(channel.find("t4Timeouts").text)

            labels = [channel_id.zfill(2)]
            us_frequency.add_metric(labels, frequency)
            us_power_level.add_metric(labels, power_level)
            us_symbol_rate.add_metric(labels, symbol_rate)
            us_timeouts.add_metric(labels + ["T1"], t1_timeouts)
            us_timeouts.add_metric(labels + ["T2"], t2_timeouts)
            us_timeouts.add_metric(labels + ["T3"], t3_timeouts)
            us_timeouts.add_metric(labels + ["T4"], t4_timeouts)
        yield from [us_frequency, us_power_level, us_symbol_rate, us_timeouts]


class LanUserExtractor(XmlMetricsExtractor):
    def __init__(self):
        super(LanUserExtractor, self).__init__(GET.LANUSERTABLE)

    def extract(self, raw_xmls: List[bytes]) -> Iterable[Metric]:
        assert len(raw_xmls) == 1
        lan_user_xml = raw_xmls[0]

        MAC_ADDRESS = "mac_address"
        HOSTNAME = "hostname"
        IPV4_ADDRESS = "ipv4_address"
        IPV6_ADDRESS = "ipv6_address"
        lan_user_speed = GaugeMetricFamily(
            "connectbox_lan_user_speed",
            "LAN user network speed",
            labels=[MAC_ADDRESS, IPV4_ADDRESS, IPV6_ADDRESS, HOSTNAME],
            unit="mbit",
        )
        root = etree.fromstring(lan_user_xml, parser=self.parser)
        assert root.tag == "LanUserTable"

        for client in root.find("Ethernet").findall("clientinfo"):
            mac_address = client.find("MACAddr").text
            ipv4_address = client.find("IPv4Addr").text
            ipv6_address = client.find("IPv6Addr").text
            hostname = client.find("hostname").text
            speed = int(client.find("speed").text)
            lan_user_speed.add_metric(
                [mac_address, ipv4_address, ipv6_address, hostname], speed
            )

        # TODO there could be more things to be reported about wifi devices here

        yield lan_user_speed


class CmStateExtractor(XmlMetricsExtractor):
    def __init__(self):
        super(CmStateExtractor, self).__init__(GET.CMSTATE)

    def extract(self, raw_xmls: List[bytes]) -> Iterable[Metric]:
        assert len(raw_xmls) == 1
        cmstate_xml = raw_xmls[0]

        root = etree.fromstring(cmstate_xml, parser=self.parser)
        assert root.tag == "cmstate"

        fahrenheit_to_celsius = lambda f: (f - 32) * 5.0 / 9
        tuner_temperature = fahrenheit_to_celsius(
            float(root.find("TunnerTemperature").text)
        )
        temperature = fahrenheit_to_celsius(float(root.find("Temperature").text))

        yield GaugeMetricFamily(
            "connectbox_tuner_temperature",
            "Tuner temperature",
            unit="celsius",
            value=tuner_temperature,
        )
        yield GaugeMetricFamily(
            "connectbox_temperature", "Temperature", unit="celsius", value=temperature,
        )


class ProvisioningStatus(Enum):
    ONLINE = "Online"
    PARTIAL_SERVICE_US = "Partial Service (US only)"
    PARTIAL_SERVICE_DS = "Partial Service (DS only)"
    PARTIAL_SERVICE_USDS = "Partial Service (US+DS)"


class DeviceStatusExtractor(XmlMetricsExtractor):
    def __init__(self):
        super(DeviceStatusExtractor, self).__init__(
            [GET.GLOBALSETTINGS, GET.CM_SYSTEM_INFO, GET.CMSTATUS]
        )

    def extract(self, raw_xmls: List[bytes]) -> Iterable[Metric]:
        assert len(raw_xmls) == 3
        global_settings_xml = raw_xmls[0]
        cm_system_info_xml = raw_xmls[1]
        cm_status_xml = raw_xmls[2]

        # parse GlobalSettings
        root = etree.fromstring(global_settings_xml, parser=self.parser)
        assert root.tag == "GlobalSettings"
        firmware_version = root.find("SwVersion").text
        cm_provision_mode = root.find("CmProvisionMode").text
        gw_provision_mode = root.find("GwProvisionMode").text
        operator_id = root.find("OperatorId").text

        # parse cm_system_info
        root = etree.fromstring(cm_system_info_xml, parser=self.parser)
        assert root.tag == "cm_system_info"
        docsis_mode = root.find("cm_docsis_mode").text
        hardware_version = root.find("cm_hardware_version").text
        uptime_as_str = root.find("cm_system_uptime").text

        # parse cmstatus
        root = etree.fromstring(cm_status_xml, parser=self.parser)
        assert root.tag == "cmstatus"
        cable_modem_status = root.find("cm_comment").text
        provisioning_status = root.find("provisioning_st").text

        yield InfoMetricFamily(
            "connectbox_device",
            "Assorted device information",
            value={
                "hardware_version": hardware_version,
                "firmware_version": firmware_version,
                "docsis_mode": docsis_mode,
                "cm_provision_mode": cm_provision_mode,
                "gw_provision_mode": gw_provision_mode,
                "cable_modem_status": cable_modem_status,
                "operator_id": operator_id,
            },
        )

        # return an enum-style metric for the provisioning status
        try:
            enum_provisioning_status = ProvisioningStatus(provisioning_status)
        except:
            raise ValueError(
                f"Unknown provisioning status '{provisioning_status}'. Please open an issue on Github."
            )
        yield StateSetMetricFamily(
            "connectbox_provisioning_status",
            "Provisioning status description",
            value={
                state.value: state == enum_provisioning_status
                for state in ProvisioningStatus
            },
        )

        # uptime is reported in a format like "36day(s)15h:24m:58s" which needs parsing
        uptime_pattern = r"(\d+)day\(s\)(\d+)h:(\d+)m:(\d+)s"
        m = re.fullmatch(uptime_pattern, uptime_as_str)
        if m is None:
            raise ValueError(
                f"Unexpected duration format '{uptime_as_str}', please open an issue on github."
            )
        uptime_timedelta = timedelta(
            days=int(m[1]), hours=int(m[2]), minutes=int(m[3]), seconds=int(m[4])
        )
        uptime_seconds = uptime_timedelta.total_seconds()

        yield CounterMetricFamily(
            "connectbox_uptime",
            "Device uptime in seconds",
            unit="seconds",
            value=uptime_seconds,
        )



from pathlib import Path
import deepmerge
from typing import Union, Dict

from ruamel.yaml import YAML

IP_ADDRESS = "ip_address"
PASSWORD = "password"
EXPORTER = "exporter"
PORT = "port"
TIMEOUT_SECONDS = "timeout_seconds"

# pick default timeout one second less than the default prometheus timeout of 10s
DEFAULT_CONFIG = {EXPORTER: {PORT: 9705, TIMEOUT_SECONDS: 9}}


def load_config(config_file: Union[str, Path]) -> Dict:
    """
    Loads and validates YAML config for this exporter and fills in default values
    :param config_file:
    :return: config as dictionary
    """
    yaml = YAML()
    with open(config_file) as f:
        config = yaml.load(f)

    # merge with defaults
    config = deepmerge.always_merger.merge(DEFAULT_CONFIG, config)

    for param in [IP_ADDRESS, PASSWORD]:
        if not param in config:
            raise ValueError(
                f"'{param}' is a mandatory config parameter, but it is missing in the YAML configuration file. Please see README.md for an example."
            )
    if config[EXPORTER][TIMEOUT_SECONDS] <= 0:
        raise ValueError(f"'{TIMEOUT_SECONDS} must be positive.")
    if config[EXPORTER][PORT] < 0 or config[EXPORTER][PORT] > 65535:
        raise ValueError(f"Invalid exporter port.")

    return config



import logging
import sys
import threading
import time

import click
from compal import Compal
from prometheus_client import CollectorRegistry, MetricsHandler
from prometheus_client.core import GaugeMetricFamily
from prometheus_client.exposition import _ThreadingSimpleServer
from requests import Timeout

from connectbox_exporter.config import (
    load_config,
    IP_ADDRESS,
    PASSWORD,
    EXPORTER,
    PORT,
    TIMEOUT_SECONDS,
)
from connectbox_exporter.xml2metric import (
    CmStateExtractor,
    LanUserExtractor,
    UpstreamStatusExtractor,
    DownstreamStatusExtractor,
    DeviceStatusExtractor,
)


class ConnectBoxCollector(object):
    def __init__(
        self, logger: logging.Logger, ip_address: str, password: str, timeout: float
    ):
        self.logger = logger
        self.ip_address = ip_address
        self.password = password
        self.timeout = timeout
        self.metric_extractors = [
            DeviceStatusExtractor(),
            CmStateExtractor(),
            LanUserExtractor(),
            DownstreamStatusExtractor(),
            UpstreamStatusExtractor(),
        ]

    def collect(self):
        successful_scrape = True
        pre_scrape_time = time.time()

        # attempt login
        try:
            self.logger.debug("Logging in at " + self.ip_address)
            connectbox = Compal(
                self.ip_address, key=self.password, timeout=self.timeout
            )
            connectbox.login()
        except (ConnectionError, Timeout, ValueError) as e:
            self.logger.error(e)
            connectbox = None
            successful_scrape = False

        # skip extracting further metrics if login failed
        if connectbox is not None:
            try:
                for extractor in self.metric_extractors:
                    contents = []
                    for fun in extractor.functions:
                        self.logger.debug(f"Querying fun=" + str(fun) + "...")
                        contents.append(connectbox.xml_getter(fun, {}).content)
                    yield from extractor.extract(contents)
            except Exception as e:
                # bail out on any error
                self.logger.error(e)
                successful_scrape = False
            finally:
                # attempt logout in case of errors
                try:
                    self.logger.debug("Logging out.")
                    connectbox.logout()
                except Exception as e:
                    self.logger.error(e)
                    successful_scrape = False

        post_scrape_time = time.time()
        if successful_scrape:
            self.logger.debug("Scrape successful.")
            yield GaugeMetricFamily(
                "connectbox_scrape_duration",
                documentation="Connect Box exporter scrape duration",
                unit="seconds",
                value=post_scrape_time - pre_scrape_time,
            )

        yield GaugeMetricFamily(
            "connectbox_scrape_success",
            documentation="Connect Box exporter scrape success",
            value=int(successful_scrape),
        )


@click.command()
@click.argument("config_file", type=click.Path(exists=True, dir_okay=False))
@click.option("-v", "--verbose", help="Print more log messages", is_flag=True)
def main(config_file, verbose):
    """
    Launch the exporter using a YAML config file.
    """
    # logging setup
    log_level = logging.DEBUG if verbose else logging.INFO
    logger = logging.getLogger("connectbox_exporter")
    logger.setLevel(log_level)
    handler = logging.StreamHandler(sys.stdout)
    handler.setLevel(log_level)
    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )
    handler.setFormatter(formatter)
    logger.addHandler(handler)

    # load user and merge with defaults
    config = load_config(config_file)
    exporter_config = config[EXPORTER]

    # fire up collector
    reg = CollectorRegistry()
    reg.register(
        ConnectBoxCollector(
            logger,
            ip_address=config[IP_ADDRESS],
            password=config[PASSWORD],
            timeout=exporter_config[TIMEOUT_SECONDS],
        )
    )

    # start http server
    CustomMetricsHandler = MetricsHandler.factory(reg)
    httpd = _ThreadingSimpleServer(("", exporter_config[PORT]), CustomMetricsHandler)
    httpd_thread = threading.Thread(target=httpd.serve_forever)
    httpd_thread.start()

    logger.info(
        f"Exporter running at http://localhost:{exporter_config[PORT]}, querying {config[IP_ADDRESS]}"
    )

    # wait indefinitely
    try:
        while True:
            time.sleep(3)
    except KeyboardInterrupt:
        httpd.shutdown()
        httpd_thread.join()



from pathlib import Path
from setuptools import find_packages, setup

with open("README.md", "rb") as f:
    long_descr = f.read().decode("utf-8")

with (Path("requirements") / "production.txt").open() as f:
    install_requires = [s.strip() for s in f.readlines()]

setup(
    name="connectbox-prometheus",
    version="0.2.0",
    author="Michael Bugert",
    author_email="git@mbugert.de",
    description='Prometheus exporter for Compal CH7465LG cable modems, commonly sold as "Connect Box"',
    long_description=long_descr,
    long_description_content_type="text/markdown",
    url="https://github.com/mbugert/connectbox-prometheus",
    entry_points={
        "console_scripts": [
            "connectbox_exporter = connectbox_exporter.connectbox_exporter:main"
        ]
    },
    packages=find_packages(exclude=["tests"]),
    install_requires=install_requires,
    include_package_data=True,
    python_requires=">=3.7",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: System :: Networking :: Monitoring",
    ],
)
